"""VQE computations steps."""

import os
import time

import numpy
from pandera import Check, Column, DataFrameSchema

from qstone.apps.computation import Computation
from qstone.connectors import connector
from qstone.utils.utils import ComputationStep, trace


class VQE(Computation):
    """
    VQE computation class.
    """

    COMPUTATION_NAME = "VQE"
    CFG_STRING = """
    {
      "cfg":
      {
        "num_required_qubits" : 4,
        "repetitions" : 10,
        "iterations" : 10,
        "compute_duration" : 1
      }
    }
    """
    SCHEMA = DataFrameSchema(
        {
            "repetitions": Column(int, Check(lambda s: s >= 0)),
            "iterations": Column(int, Check(lambda s: s >= 0)),
            "compute_duration": Column(int, Check(lambda s: s >= 0)),
        }
    )

    def __init__(self, cfg: dict):
        super().__init__(cfg)
        # Hints - do not remove
        self.repetitions: int
        self.iterations: int
        self.compute_duration: int
        self.num_shots = int(os.environ.get("NUM_SHOTS", self.repetitions))

    def _generate_circuit(self, datapath: str, num_qubits: int):
        """Generate a random (small) quantum circuit"""
        qasm = 'OPENQASM 2.0;\ninclude "qelib1.inc";'
        qasm += f"\nqreg q[{num_qubits}];\ncreg c[{num_qubits}];"
        # Add N rounds of gates.
        for i in range(numpy.random.randint(4, 20)):
            qasm += numpy.random.choice(
                [
                    f"rx({i/4}) q[{i%num_qubits}];\n",
                    f"rz({i/3}) q[{i%num_qubits}];\n",
                    f"h q[{i%num_qubits}];\n",
                ]
            )
        # Reading all qubit states
        for i in range(num_qubits):
            qasm += f"measure q[{i}] -> c[{i}];\n"
        # Write it back
        with open(
            f"{os.path.join(datapath, 'random')}.qasm", "w", encoding="utf-8"
        ) as fid:
            fid.write(qasm)
        print(f"Generated VQE-like circuit with {num_qubits} qubits")

    def _mock_compute(self):
        """Function that mimics computation delay. Simple sleep of constant duration."""
        time.sleep(self.compute_duration / 1000)

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.PRE)
    def pre(self, datapath: str):
        """Phase not required"""

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.RUN)
    def run(self, datapath: str, connection: connector.Connector):
        """Runs the Quantum circuit N times"""
        # Executing self.iterations iterations with a maximum of self.repetitions shots each
        for _ in range(numpy.random.randint(1, self.iterations)):
            self._generate_circuit(datapath, int(os.environ["NUM_QUBITS"]))
            connection.run(
                qasm=f"{os.path.join(datapath, 'random')}.qasm", reps=self.num_shots
            )
            self._mock_compute()

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.POST)
    def post(self, datapath: str):
        """Phase not required"""



"""
Generation of the testbench.
"""

import argparse
import base64
import math
import os
import pickle
import shutil
import tarfile
from typing import Any, List, Union

import numpy
import pandas as pa
from jinja2 import Template

from qstone.utils.utils import QpuConfiguration, parse_json

SCHEDULERS = {
    "bare_metal": "bare_metal",
    "jsrun": "lsf/jsrun",
    "slurm": "slurm/schedmd",
}
SCHEDULER_CMDS = {"bare_metal": "bash", "jsrun": "jrun", "slurm": "sbatch"}
SCHEDULER_EXTS = {"bare_metal": "sh", "jsrun": "bsub", "slurm": "sbatch"}

SCHEDULER_ARGS = {"walltime": "3", "nthreads": "1"}

CURRENT_PATH = os.path.dirname(os.path.realpath(__file__))
GEN_PATH = "qstone_suite"


def _check_nan(val):
    return isinstance(val, float) and (numpy.isnan(val) or math.isnan(val))


def _get_value(job_cfg: pa.DataFrame, key: str, default: str):
    val = default
    try:
        v = job_cfg[key]
        val = v if isinstance(v, float) else v.values[0]
    except (KeyError, IndexError):
        pass
    # Check for both numpy.nan and Python's float nan
    if _check_nan(val):
        val = default
    return str(val)


def _find_files(sched_path: str):
    search_paths = [sched_path, os.path.join(CURRENT_PATH, "common")]
    all_files = [
        os.path.join(search_path, s)
        for search_path in search_paths
        for s in os.listdir(search_path)
        if s not in ["__pycache__", ".cache"] and not s.endswith(".pyc")
    ]
    jinja_files = [s for s in all_files if s.endswith("jinja")]
    non_jinja_files = list(set(all_files) - set(jinja_files))
    return (jinja_files, non_jinja_files)


def _render_templates(
    sched: str,
    sched_path: str,
    subs: dict,
    job_types: List[str],
    jobs_cfg: pa.DataFrame,
):
    """Convert all templates and add all the files that are in the scheduler folder"""
    # Ignore folders and search in the search paths all the paths
    jinja_files, non_jinja_files = _find_files(sched_path)
    # Adding templated files
    for jinja_file in jinja_files:
        with open(jinja_file, encoding="utf-8") as fid:
            source = fid.read()
        if "{app}" in jinja_file:
            for t in job_types:
                outfile = os.path.join(
                    GEN_PATH,
                    os.path.basename(
                        jinja_file.replace(".jinja", "").replace("{app}", t)
                    ),
                )
                j = jobs_cfg[jobs_cfg["type"] == t]
                args = {
                    key: _get_value(j, key, val) for key, val in SCHEDULER_ARGS.items()
                }
                sched_args = {"sched_args": _get_value(j, f"{sched}_opt", "")}
                Template(source).stream({**subs, **args, **sched_args}).dump(outfile)
        else:
            outfile = os.path.join(
                GEN_PATH, os.path.basename(jinja_file.replace(".jinja", ""))
            )
            Template(source).stream(subs).dump(outfile)
    # Adding non template files
    for non_jinja_file in non_jinja_files:
        shutil.copy(
            non_jinja_file,
            f"{os.path.join(GEN_PATH, os.path.basename(non_jinja_file))}",
        )


def _render_and_pack(
    scheduler: str,
    output_filename: str,
    subs: dict,
    job_types: List[str],
    jobs_cfg: pa.DataFrame,
):
    """
    Renders and packs all the necessary files to run as a user
    """
    sched = SCHEDULERS[scheduler]
    sched_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), sched)
    shutil.rmtree(GEN_PATH, ignore_errors=True)
    os.makedirs(GEN_PATH)
    _render_templates(sched, sched_path, subs, job_types, jobs_cfg)
    # Copy the required files.
    with tarfile.open(output_filename, "w:gz") as tar:
        # Adding necessary scripts excluding original templates
        tar.add(GEN_PATH, recursive=True)
        for job_type in job_types:
            # Adding user defined apps
            job_cfg = jobs_cfg[jobs_cfg["type"] == job_type]
            app = _get_value(job_cfg, "path", "")
            if app:
                assert os.path.exists(app)
                tar.add(
                    app,
                    arcname=f"{GEN_PATH}/{os.path.basename(app)}",
                    recursive=False,
                )
    shutil.rmtree(GEN_PATH)


def _compute_job_pdf(usr_cfg: "pa.Series[Any]") -> List[float]:
    """Computes the normalized pdf to assign to different jobs based on user
    configurations and speciified qubit capacity
    """

    pdf = [prob for comp, prob in usr_cfg["computations"].items()]

    normalized = [float(p) / sum(pdf) for p in pdf]

    return normalized


def _randomise(vals, def_val):
    """Return randomised value from range when available"""

    if pa.isnull(vals).any():
        value = def_val
    else:
        values = vals.tolist()[0]
        if len(values) > 1:
            value = numpy.random.randint(values[0], values[1])
        else:
            value = values[0]
    return value


def _to_bytes(ob):
    return base64.b64encode(pickle.dumps(ob)).decode("utf-8")


def _generate_user_jobs(
    usr_cfg: "pa.Series[Any]",
    jobs_cfg: pa.DataFrame,
    job_pdf: List[float],
    job_count: int,
):
    """
    Generates the different user jobs provided given the configuration and the number of
    calls.
    """
    runner = 'python "$EXEC_PATH"/type_exec.py'
    job_types = numpy.random.choice(
        list(usr_cfg["computations"].keys()), p=job_pdf, size=(job_count)
    )
    # Check that we have generated a not empty
    assert (
        len(job_types) > 0
    ), "Configuration generated zero jobs. Please check your configuration file."

    # Randomise number of qubits
    num_qubits = []
    num_shots = []
    app_logging_level = []
    app_args = []

    def_qubits = 2
    def_shots = 100
    for j in job_types:
        app_cfg = jobs_cfg[jobs_cfg["type"] == j]
        if app_cfg.empty:
            num_qubits.append(def_qubits)
            num_shots.append(def_shots)
            app_args.append("")
            app_logging_level.append(2)
        else:
            num_qubits.append(_randomise(app_cfg["qubits"], def_qubits))
            num_shots.append(_randomise(app_cfg["num_shots"], def_shots))
            if "app_args" in app_cfg.columns:
                t = app_cfg["app_args"].tolist()[0]
                if not _check_nan(t):
                    app_args.append(_to_bytes(t))
            if "app_logging_level" in app_cfg.columns:
                t = app_cfg["app_logging_level"].tolist()[0]
                if not _check_nan(t):
                    app_logging_level.append(int(t))
    # Assign job id and pack
    job_ids = list(range(len(job_types)))
    return (
        list(
            zip(
                [f"{runner} {s}" for s in job_types],
                num_qubits,
                job_ids,
                num_shots,
                app_logging_level,
                app_args,
            )
        ),
        set(job_types),
    )


def _environment_variables_exports(env_vars: dict) -> List[str]:
    """
    Generates export statements for environment variables, handling nested dictionaries.

    For nested dictionaries like {"a": {"b": x, "c": y}}, generates symbols:
    A_B = x, A_C = y

    Args:
        env_vars: Dictionary of environment variables, potentially nested

    Returns:
        List of export statements for all environment variables
    """
    exports_list = []

    def process_dict(current_dict, prefix=""):
        for key, value in current_dict.items():
            # Convert the current key to uppercase
            upper_key = key.upper().replace(".", "_")

            # If there's a prefix, we're in a nested structure
            if prefix:
                current_prefix = f"{prefix}_{upper_key}"
            else:
                current_prefix = upper_key

            # If value is a dictionary, process it recursively
            if isinstance(value, dict):
                process_dict(value, current_prefix)
            else:
                # Create export statement for leaf values
                exports_list.append(f'export {current_prefix}="{value}"')

    # Start processing from the root dictionary
    process_dict(env_vars)
    return exports_list


def generate_suite(
    config: str, job_count: int, output_folder: str, atomic: bool, scheduler: str
) -> List[str]:
    """
    Generates the suites of jobs for the required users.

    Args:
        config: Input configuration for generate, defines QPU configuration and user jobs
        job_count: Number of jobs to generate per user
        output_folder: Scheduler tar file output location
        atomic: optional flag to create a single job out of the three phase
        scheduler: target HPC scheduler

    Returns list of output file paths
    """
    # Get configurations
    config_dict = parse_json(config)
    env_cfg = config_dict["environment"]
    users_cfg = pa.DataFrame(config_dict["users"])
    jobs_cfg = pa.DataFrame(config_dict["jobs"])

    env_exports = _environment_variables_exports(env_cfg)

    qpu_config = QpuConfiguration()
    qpu_config.load_configuration(env_cfg)

    if not job_count:
        job_count = env_cfg["job_count"]

    # Generating list of jobs
    output_paths = []
    for prog_id, user_cfg in users_cfg.iterrows():
        pdf = _compute_job_pdf(user_cfg)
        # Get the job count either from global or user configuration.
        job_count_user = float(
            _get_value(user_cfg.to_frame(), "job_count", str(job_count))
        )
        jobs, job_types = _generate_user_jobs(
            user_cfg, jobs_cfg, pdf, int(job_count_user)
        )

        # generate substitutions for Jinja templates
        formatted_jobs = [" ".join(map(str, job)) for job in jobs]

        user_name = user_cfg["user"]
        usr_env_exports = [
            f'export PROG_ID="{prog_id}"',
            f'export QS_USER="{user_name}"',
        ]
        subs = {
            "exports": "\n".join(env_exports + usr_env_exports),
            "jobs": "\n".join(formatted_jobs),
            "project_name": env_cfg["project_name"],
            "atomic": atomic,
            "sched_ext": SCHEDULER_EXTS[scheduler],
            "sched_cmd": SCHEDULER_CMDS[scheduler],
            "sched_aware": (
                "--gres=qpu:1" if env_cfg["scheduling_mode"] == "SCHEDULER" else ""
            ),
        }
        # Pack project files
        filename = os.path.join(output_folder, f"{scheduler}_{user_name}.qstone.tar.gz")
        # render and pack all the files
        _render_and_pack(scheduler, filename, subs, job_types, jobs_cfg)
        output_paths.append(filename)
    return output_paths


def main():
    """
    Runs the generator phase.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("config", type=str)
    parser.add_argument("job_count", type=int)
    parser.add_argument("output_folder", type=str)
    parser.add_argument("scheduler", type=str, choices=SCHEDULERS)
    parser.add_argument("atomic", type=bool, action="store_true")
    args = parser.parse_args()
    generate_suite(
        args.config, args.job_count, args.output_folder, args.atomic, args.scheduler
    )


if __name__ == "__main__":
    main()



"""RB computations steps."""

import os

import numpy as np
import pygsti
from pandera import Check, Column, DataFrameSchema
from pygsti.processors import CliffordCompilationRules as CCR
from pygsti.processors import QubitProcessorSpec as QPS

from qstone.apps.computation import Computation
from qstone.connectors import connector
from qstone.utils.utils import ComputationStep, trace


class RB(Computation):
    """
    Type 1 computation class.
    """

    COMPUTATION_NAME = "RB"
    CFG_STRING = """
    {
      "cfg":
      {
         "num_required_qubits" : 2,
         "repetitions": 2
      }
    } 
    """
    SCHEMA = DataFrameSchema(
        {
            "repetitions": Column(int, Check(lambda s: s >= 0)),
        }
    )

    def __init__(self, cfg: dict):
        super().__init__(cfg)
        # Hints - do not remove
        self.repetitions: int
        self.num_shots = int(os.environ.get("NUM_SHOTS", self.repetitions))

    def _get_configuration(self):
        """Returns standard configuration for the job."""
        n_qubits = int(os.environ["NUM_QUBITS"])
        if n_qubits > 2:
            n_qubits = 2
            print("Maximum number of supported qubits is 2. Trimming down request")
        n_active_qubits = 1
        depths = [1, 4, 8, 16]
        reps = 10
        qubit_labels = [f"Q{s}" for s in range(n_qubits)]
        gate_names = ["Gxpi2", "Gxmpi2", "Gypi2", "Gympi2", "Gcphase"]
        availability = {
            "Gcphase": [
                ("Q0", "Q1"),
                ("Q1", "Q0"),
            ]
        }
        pspec = QPS(
            n_qubits, gate_names, availability=availability, qubit_labels=qubit_labels
        )
        compilations = {
            "absolute": CCR.create_standard(
                pspec, "absolute", ("paulis", "1Qcliffords"), verbosity=0
            ),
            "paulieq": CCR.create_standard(
                pspec, "paulieq", ("1Qcliffords", "allcnots"), verbosity=0
            ),
        }
        # Only using Q0 for now
        active_qubit_labels = ["Q0"]
        design = pygsti.protocols.CliffordRBDesign(
            pspec, compilations, depths, reps, active_qubit_labels
        )
        print(f"Generated benchmarking routing with {n_qubits} qubits")
        return (design, pspec, active_qubit_labels, n_active_qubits)

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.PRE)
    def pre(self, datapath: str):
        """
        Generates the circuits required to extract the randomised benchmarking
        circuits.
        """
        run_file = f"{datapath}/RB_run_{os.environ['JOB_ID']}.npz"
        design, _, active_qubit_labels, n_active_qubits = self._get_configuration()
        # Translation to OpenQASM
        circuit_list_nested = design.all_circuits_needing_data
        qubit_indices = {label: i for i, label in enumerate(active_qubit_labels)}
        qasms = [
            c.convert_to_openqasm(
                num_qubits=n_active_qubits,
                standard_gates_version="x-sx-rz",
                qubit_conversion=qubit_indices,
            )
            for c in circuit_list_nested
        ]
        # Flatten out fully. We might want to consider keeping the N steps separate
        exp = list(np.concatenate(design.idealout_lists).flat)
        # Generate a .npz file that contains the configuration
        np.savez(run_file, qasms=qasms, exp=exp, allow_pickle=True)

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.RUN)
    def run(self, datapath: str, connection: connector.Connector):
        """Runs the Quantum circuit N times

        Args:
            datapath: path location to write circuit
            connection: connector object to run circuit
            shots: number of shots to be executed
        """
        run_file = f"{datapath}/RB_run_{os.environ['JOB_ID']}.npz"
        # Find number of circuits
        results = []
        # load configuration (and qasm files)
        vals = np.load(run_file)
        i = 0
        for circuit in vals["qasms"]:
            path = os.path.join(
                os.path.dirname(run_file), f"RB_{os.environ['JOB_ID']}_{str(i)}.qasm"
            )
            with open(path, "w", encoding="utf-8") as fid:
                fid.write(circuit)
            i = i + 1
            results.append(connection.run(qasm=path, reps=self.num_shots))
        store = {"qasms": vals["qasms"], "exp": vals["exp"], "res": results}
        np.savez(run_file, **store, allow_pickle=True)

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.POST)
    def post(self, datapath: str):
        """Runs the processing of the randomised benchmarking using the data provided by
        datapath
        """
        run_file = f"{datapath}/RB_run_{os.environ['JOB_ID']}.npz"
        report_file = f"{datapath}/RB_report_{os.environ['JOB_ID']}.txt"
        vals = np.load(run_file, allow_pickle=True)
        exp = vals["exp"]
        res = vals["res"]
        success_counts = np.count_nonzero(exp == res)
        with open(report_file, "w", encoding="utf-8") as fid:
            fid.write(f"success_counts: {success_counts}")



"""Main file for qstone."""

import argparse
import logging
import os
import subprocess
from typing import Optional, Sequence

from qstone.generators import generator
from qstone.profiling import profile


def generate(args: Optional[Sequence[str]] = None) -> None:
    """Qstone cli subcommand for generator."""
    logger = logging.getLogger()
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    logger.info("Generating benchmark tarballs")

    # Adjusting job count
    job_count_s = args.job_count  # type: ignore[union-attr]
    if job_count_s is not None:
        job_count = int(job_count_s)
    else:
        job_count = None

    generated_files = generator.generate_suite(
        config=args.src,  # type: ignore[union-attr]
        job_count=job_count,  # type: ignore[arg-type]
        output_folder=args.dst,  # type: ignore[union-attr]
        atomic=args.atomic,  # type: ignore[union-attr]
        scheduler=args.scheduler,  # type: ignore[union-attr]
    )

    logger.info("Generated %s tar balls:", len(generated_files))
    logger.info("\n".join(generated_files))


def run(args: Optional[Sequence[str]] = None) -> None:
    """Qstone cli subcommand for executing scheduler."""
    logger = logging.getLogger()
    logging.basicConfig(level=logging.INFO, format="%(message)s")

    if args.dst:  # type: ignore[union-attr]
        destination = args.dst  # type: ignore[union-attr]
        os.makedirs(destination, exist_ok=True)
    else:
        destination = "."
    logger.info(
        ">>Extracting benchmark in %s", os.path.join(destination, "qstone_suite")
    )
    result = subprocess.run(
        ["tar", "-xvf", args.src, "-C", destination],  # type: ignore[union-attr]
        check=False,
        stdout=subprocess.DEVNULL,
    )
    logger.info(">>Starting benchmark... ")

    result = subprocess.run(
        ["sh", os.path.join(destination, "qstone_suite", "qstone.sh")], check=False
    )
    logger.info("Scheduler ran with status %s", result.returncode)


def prof(args: Optional[Sequence[str]] = None) -> None:
    """Qstone cli subcommand for profiling scheduler exection."""
    logger = logging.getLogger()
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    logger.info("Extracting scheduler tar file")
    # Folder can be a list of folders
    f = [args.folder] if isinstance(args.folder, str) else args.folder  # type: ignore[union-attr]
    profile.profile(args.cfg, f, args.pickle)  # type: ignore[union-attr]


def main(arg_strings: Optional[Sequence[str]] = None) -> None:
    """Qstone main entry point function.

    Args:
        arg_strings: Optionally manually provide argument strings (only used in tests).
    """
    parser = argparse.ArgumentParser(prog="qstone")

    subparsers = parser.add_subparsers(required=True)

    gen_cmd = subparsers.add_parser("generate", help="Generates job scheduler")

    gen_cmd.add_argument(
        "-i",
        "--src",
        help="Path to the input configuration",
        required=True,
        type=str,
    )

    gen_cmd.add_argument(
        "-o",
        "--dst",
        help="Path to the folder where the schedulers will be generated ",
        default=".",
        required=False,
        type=str,
    )

    gen_cmd.add_argument(
        "-n",
        "--job_count",
        help="Number of jobs to generate",
        default=None,
        required=False,
        type=str,
    )

    gen_cmd.add_argument(
        "-a",
        "--atomic",
        help="Generate a single job per application, default is three [pre,run,post]",
        default=False,
        required=False,
        action="store_true",
    )

    gen_cmd.add_argument(
        "-s",
        "--scheduler",
        help="Generate the configuration for a specific scheduler",
        default="bare_metal",
        choices=["slurm", "jsrun", "bare_metal"],
        required=False,
        type=str,
    )

    gen_cmd.set_defaults(func=generate)

    runner = subparsers.add_parser("run", help="Run scheduler")

    runner.add_argument(
        "-i",
        "--src",
        help="Path to scheduler tar file",
        required=True,
        type=str,
    )
    runner.add_argument(
        "-o",
        "--dst",
        help="Path to extract scheduler to ",
        required=False,
        type=str,
    )

    runner.set_defaults(func=run)

    profiler = subparsers.add_parser("profile", help="Profile job execution")

    profiler.add_argument(
        "--cfg", type=str, help="Configuration file used to generate the load"
    )
    profiler.add_argument(
        "--folder",
        type=str,
        action="append",
        help="Folder that contains the runs, repeatable argument",
    )
    profiler.add_argument(
        "--pickle",
        type=str,
        help="Optional Pickle filepath to store pickled dataframe",
        default="./QS_Profile.pkl",
    )

    profiler.set_defaults(func=prof)

    args = parser.parse_args(arg_strings)
    args.func(args)


if __name__ == "__main__":
    main()



"""QPU computation class and configuration dataclass"""

import base64
import json
import os
import pickle
from abc import ABC, abstractmethod
from typing import Dict, Optional

import pandas as pd
from pandera import DataFrameSchema

from qstone.connectors import connector
from qstone.utils.utils import QpuConfiguration


def byte_to_dict(s: str) -> Dict:
    """converts a json formatted string into a dictionary"""
    if not s:
        return {}
    decoded = base64.b64decode(s.encode("utf-8"))
    return pickle.loads(decoded)


class Computation(ABC):
    """
    Abstract class for a QPU computation.
    To be ovverriden  by implemented jobs.

    Args:
        cfg: Computation configuration dictionary. Key value pairs describing
        the script data.
    """

    COMPUTATION_NAME: str
    CFG_PATH: Optional[str] = None
    CFG_STRING: Optional[str] = None
    SCHEMA: DataFrameSchema
    BASEPATH: str = os.path.join(os.path.dirname(__file__), "..")

    def __init__(self, cfg: Dict):
        self._cfg = cfg
        for key, val in cfg.items():
            setattr(self, key, val)
        self._qpu_cfg = QpuConfiguration()
        self._app_args = byte_to_dict(os.environ.get("APP_ARGS", ""))
        self._logging_level = os.environ.get("LOGGING_LEVEL", "")

    @classmethod
    def from_json(cls, path: Optional[str] = None):
        """Factory method for creating computation via JSON configuration.

        Args:
            path: Path to computation configuration file
        """
        if cls.CFG_STRING is None:
            if path is None:
                path = os.path.join(cls.BASEPATH, cls.CFG_PATH)  # type:ignore
            with open(path, encoding="utf-8") as fo:
                json_obj = json.load(fo)
        else:
            json_obj = json.loads(cls.CFG_STRING)
        cfg = json_obj["cfg"]
        df = pd.json_normalize(cfg)
        cls.SCHEMA.validate(df)
        return cls(cfg)

    def dump_cfg(self) -> str:
        """
        Serializes computation script information into JSON string.

        Returns json string representation of computation configuration.
        """
        return json.dumps(self._cfg)

    @property
    def qpu_cfg(self):
        """qpu_cfg getter"""
        return self._qpu_cfg

    @abstractmethod
    def pre(self, datapath: str) -> None:
        """QPU computation preprocessing step"""
        raise NotImplementedError

    @abstractmethod
    def run(self, datapath: str, connection: connector.Connector):
        """QPU computation circuit run step"""
        raise NotImplementedError

    @abstractmethod
    def post(self, datapath):
        """QPU computation postprocessing step"""
        raise NotImplementedError



"""Connectors for different Quantum Stacks"""

import warnings
from enum import Enum
from typing import Optional

from qstone.connectors import connection

try:
    from qstone.connectors.grpc.runner import GRPCConnecction
except ImportError:
    warnings.warn("grpc failed to import", ImportWarning)
from qstone.connectors.backends.rigetti.runner import RigettiConnection
from qstone.connectors.http.runner import HttpConnection
from qstone.connectors.no_link.no_link import NoLinkConnection


class ConnectorType(Enum):
    """Type of connection"""

    GRPC = "GRPC"
    NO_LINK = "NO_LINK"
    HTTPS = "HTTPS"
    RIGETTI = "RIGETTI"


class Connector:
    """Class used to hold connection between HPC compute node and Quantum bridge"""

    def __init__(
        self,
        conn_type: ConnectorType,
        mode: str,
        qpu_host: str,
        qpu_port: int,
        compiler_host: str,
        compiler_port: int,
        target: str,
        lockfile: Optional[str],
    ):
        """Initialise the connector object"""
        self._protocol = conn_type
        self._mode = mode
        self._qpu_host = qpu_host
        self._qpu_port = qpu_port
        self._compiler_host = compiler_host
        self._compiler_port = compiler_port
        self._target = target
        self._connection: connection.Connection
        self._lockfile: Optional[str] = None if lockfile == "NONE" else lockfile

        if self.protocol == ConnectorType.GRPC:
            self._connection = GRPCConnecction()
        elif self.protocol == ConnectorType.HTTPS:
            self._connection = HttpConnection()
        elif self.protocol == ConnectorType.NO_LINK:
            self._connection = NoLinkConnection()
        elif self.protocol == ConnectorType.RIGETTI:
            self._connection = RigettiConnection()

    @property
    def protocol(self):
        """Returns the protocol information"""
        return self._protocol

    @property
    def mode(self):
        """Returns the mode in use (EMULATION, REAL, RANDOM)"""
        return self._mode

    @property
    def qpu_host(self):
        """Returns the qpu host name"""
        return self._qpu_host

    @property
    def qpu_port(self):
        """Returns the qpu port"""
        return self._qpu_port

    @property
    def compiler_host(self):
        """Returns the compiler host name"""
        return self._compiler_host

    @property
    def compiler_port(self):
        """Returns the compiler_port"""
        return self._compiler_port

    @property
    def connection(self):
        """Returns the server_port"""
        return self._connection

    @property
    def address(self):
        """Returns the address of the Quantum bridge device"""
        return f"{self.qpu_host}::{self.qpu_port}"

    @property
    def target(self):
        """Returns the target name"""
        return self._target

    @property
    def lockfile(self):
        """Returns the lockfile information"""
        return self._lockfile

    def run(self, qasm: str, reps: int):
        """Runs the provided QASM circuit"""
        return self.connection.run(
            qasm,
            reps,
            self.mode,
            self.qpu_host,
            self.qpu_port,
            self.compiler_host,
            self.compiler_port,
            self.target,
            self.lockfile,
        )



"""Defines the two steps of PyMatching jobs"""

import os

import numpy as np
import pymatching
from pandera import Check, Column, DataFrameSchema
from stim import Circuit  # pylint:disable=no-name-in-module

from qstone.apps.computation import Computation
from qstone.connectors import connector
from qstone.utils.utils import ComputationStep, trace


class PyMatching(Computation):  # pylint:disable=invalid-name
    """
    PyMatching computation class.
    """

    COMPUTATION_NAME = "PyMatching"
    CFG_STRING = """
    {
      "cfg":
      {
        "num_required_qubits" : 10,
        "repetitions": 2
      }
    }
    """
    SCHEMA = DataFrameSchema(
        {
            "repetitions": Column(int, Check(lambda s: s >= 0)),
        }
    )

    def __init__(self, cfg: dict):
        super().__init__(cfg)
        # Hints - do not remove
        self.repetitions: int
        self.num_shots = int(os.environ.get("NUM_SHOTS", self.repetitions))

    def _initialise_circuit(self):
        """Initialise the circuit using Stim"""
        circuit = Circuit.generated(
            "surface_code:rotated_memory_x",
            distance=3,
            rounds=5,
            after_clifford_depolarization=0.005,
        )
        model = circuit.detector_error_model(decompose_errors=True)
        matching = pymatching.Matching.from_detector_error_model(model)
        sampler = circuit.compile_detector_sampler()
        return (matching, sampler)

    def _convert_stim_circuit(self, stim_circuit: Circuit):
        """Convert stim circuit to standardised circuit format

        Args:
            circuit: stim circuit object

        Returns: circuit converted from stim circuit to format
        to be used on runner, current standard is QASM
        """
        # Noise not supported in QASM
        noiseless_circuit = stim_circuit.without_noise()
        qasm_circuit = noiseless_circuit.to_qasm(open_qasm_version=3)
        return qasm_circuit

    def generate_synthetic_data(self, data_path: str):
        """Generates synthetic data and stores them in data_path"""
        _, sampler = self._initialise_circuit()
        syndrome, actual_observables = sampler.sample(
            shots=self.num_shots, separate_observables=True
        )
        np.savez(data_path, syn=syndrome, obs=actual_observables)

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.PRE)
    def pre(self, datapath: str):
        """Prepare and write circuit for QEC experiment

        Args:
            datapath: path location to write circuit

        Returns: path location of written circuit, without extension
        """
        stim_circuit = Circuit.generated(
            "surface_code:rotated_memory_x",
            distance=3,
            rounds=5,
            after_clifford_depolarization=0.005,
        ).without_noise()

        qasm_circuit = self._convert_stim_circuit(stim_circuit)
        circuit_path = os.path.join(datapath, f"PyMatching_{os.environ['JOB_ID']}")

        # Write qasm circuit
        with open(f"{circuit_path}.qasm", "w", encoding="utf-8") as fid:
            fid.write(str(qasm_circuit))

        # Write stim circuit to extract sampler on run
        with open(f"{circuit_path}.stim", "w", encoding="utf-8") as fid:
            stim_circuit.to_file(fid)

        return circuit_path

    def get_creg_indexes(self, qasm_content: str):
        """
        Parse a QASM file content to determine the indexes of classical registers (cregs)
        assuming first-come-first-serve allocation.

        Args:
            qasm_content (str): Content of the QASM file as a string

        Returns:
            dict: Dictionary mapping creg names to their bit ranges as tuples (start, end)
        """
        # Initialize variables
        creg_info = {}
        current_index = 0

        # Split the content into lines
        lines = qasm_content.strip().split("\n")

        # Process each line
        for line in lines:
            # Remove comments if any
            if "//" in line:
                line = line[: line.index("//")]

            line = line.strip()

            # Look for creg declarations
            if line.startswith("creg "):
                # Extract the creg name and size
                # Format is typically: creg name[size];
                parts = line.replace("creg ", "").replace(";", "").strip().split("[")
                if len(parts) == 2:
                    creg_name = parts[0].strip()
                    size_part = parts[1].split("]")[0].strip()

                    try:
                        size = int(size_part)
                        # Calculate the index range for this creg
                        start_index = current_index
                        end_index = current_index + size - 1

                        # Store the information
                        creg_info[creg_name] = (start_index, end_index)

                        # Update the current index for the next creg
                        current_index += size
                    except ValueError:
                        # Skip if size is not a valid integer
                        continue

        return creg_info

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.RUN)
    def run(self, datapath: str, connection: connector.Connector):
        """Runs the Quantum circuit N times

        Args:
            datapath: path location to write circuit
            connection: connector object to run circuit
            shots: number of shots to be executed

        Returns: path location of syndromes file
        """

        circuit_path = os.path.join(datapath, f"PyMatching_{os.environ['JOB_ID']}")

        # Send circuit to connector
        results = connection.run(qasm=f"{circuit_path}.qasm", reps=self.num_shots)

        # Get det and obs indexes
        with open(f"{circuit_path}.qasm", "r", encoding="utf-8") as fid:
            creg_ranges = self.get_creg_indexes(fid.read())

        # Convert syndromes to np array and write to path
        meas = results["measurements"]
        dets_idx = creg_ranges["dets"]
        obs_idx = creg_ranges["obs"]
        syndrome = np.array(meas[dets_idx[0] : dets_idx[1]], dtype=bool)
        obs = np.array(meas[obs_idx[0] : obs_idx[1]], dtype=bool)

        syndrome_path = os.path.join(
            datapath, f"PyMatching_{os.environ['JOB_ID']}_syndromes.npz"
        )

        np.savez(syndrome_path, syn=syndrome, obs=obs)

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.POST)
    def post(self, datapath: str):
        """Runs the postprocessing analysis using PyMatching decoder over a given input file

        Args:
            datapath: path location to write circuit

        Returns: number of errors detected
        """

        circuit_path = os.path.join(datapath, f"PyMatching_{os.environ['JOB_ID']}")

        stim_circuit = Circuit.from_file(f"{circuit_path}.stim")

        # Synthetic data from stim circuit
        sampler = stim_circuit.compile_detector_sampler()
        sampled_syndromes, sample_observables = sampler.sample(
            shots=1000, separate_observables=True
        )

        model = stim_circuit.detector_error_model(decompose_errors=True)
        matching = pymatching.Matching.from_detector_error_model(model)
        print(f"MARCO: datapath - {datapath}")
        syndrome_path = os.path.join(
            datapath, f"PyMatching_{os.environ['JOB_ID']}_syndromes.npz"
        )
        vals = np.load(syndrome_path)

        synd = vals["syn"]
        actual_observables = vals["obs"]

        # In case of simulating circuit by generating random readouts PyMatching will fail
        # So we use the sampled syndromes instead
        try:
            predicted_observables = matching.decode_batch(synd)
        except ValueError:
            predicted_observables = sampled_syndromes
            actual_observables = sample_observables

        num_errors = np.sum(np.any(predicted_observables != actual_observables, axis=1))

        return num_errors



"""MPI handler in case of no MPI Support - passthrough"""


class MPIHandler:
    """MPIHandler wraps MPI calls"""

    def Get_size(self):  # pylint: disable=invalid-name
        """Returns the size of the channel"""
        return 0

    def Get_rank(self):  # pylint: disable=invalid-name
        """Returns the rank of the channel"""
        return 0

    def BCast(self, **kwargs):  # pylint: disable=invalid-name
        """BCast call - mocked"""
        raise NotImplementedError("Requires MPI support")

    def allreduce(self, **kwargs):  # pylint: disable=invalid-name
        """allreduce call - mocked"""
        raise NotImplementedError("Requires MPI support")

    def allgather(
        self, sendobj=None, recvobj=None
    ):  # pylint: disable=[invalid-name, unused-argument]
        """allgather call - mocked"""
        return []

    def Barrier(self):  # pylint: disable=invalid-name
        """Barrier call - mocked"""
        pass



"""Quantum executor over a HTTP channel"""

import json
import os
import secrets
import sys
from time import sleep

import requests

from qstone.connectors import connection
from qstone.utils.utils import ComputationStep, QpuConfiguration, trace


class HttpConnection(connection.Connection):
    """Connection running jobs over Http"""

    def __init__(self):
        """Creates the empty response"""
        self.response = None
        self.http_timeout = int(os.environ.get("TIMEOUTS_HTTP", 10))
        self.lock_timeout = int(os.environ.get("TIMEOUTS_LOCK", 200))

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.PRE,
    )
    def preprocess(self, qasm_ptr: str) -> str:
        """Preprocess the data."""
        # Currently passthrough.
        with open(qasm_ptr, "r", encoding="utf-8") as fid:
            return fid.read()

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.POST,
    )
    def postprocess(self, message: str) -> str:
        """Postprocess the data"""
        # If the message is None we return an empty string.
        # Please note. In this model it is responsability of the gateway machine to
        # return the data in the correct format as defined in assumptions.md
        return json.loads(message) if message else ""

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.RUN,
        label="_request_and_process",
    )
    def _request_and_process(self, qasm_ptr: str, reps: int, hostpath: str):
        pkt_id = secrets.randbelow(2**31)
        circuit = self.preprocess(qasm_ptr)
        payload = {"circuit": circuit, "pkt_id": pkt_id, "reps": reps}
        headers: dict = {}
        r = requests.post(
            f"{hostpath}/execute", timeout=10, headers=headers, json=payload
        )
        success = r.status_code == 200
        if success:
            r = requests.get(
                f"{hostpath}/results",
                timeout=self.http_timeout,
                json={"pkt_id": pkt_id},
            )
            self.response = r.text
            success = r.status_code == 200
        if not success:
            sys.stderr.write("QSTONE::ERR - Request failed")
        return success

    # mypy: disable-error-code="attr-defined"
    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.RUN,
    )
    def run(
        self,
        qasm_ptr: str,
        reps: int,
        mode: str,
        qpu_host: str,
        qpu_port: int,
        compiler_host: str,
        compiler_port: int,
        target: str,
        lockfile: str,
    ) -> dict:
        """Run the connection to the server"""
        qpu_hostpath = f"{qpu_host}:{qpu_port}" if qpu_port else qpu_host
        # Prepending the HTTP specifier if not provided.
        qpu_hostpath = (
            qpu_hostpath
            if qpu_hostpath.startswith("http://")
            else f"http://{qpu_hostpath}"
        )

        # Active wait on lock
        if lockfile is not None:
            owned = False
            lock = connection.FileLock(lockfile)
            for _ in range(10 * self.lock_timeout):
                if lock.acquire_lock():
                    owned = True
                    break
                sleep(0.1)
            if not owned:
                sys.stderr.write("QSTONE::ERR - timeout waiting for lock")
                return {}

        self._request_and_process(qasm_ptr, reps, qpu_hostpath)

        # releasing lock takes care of None case as well.
        if lockfile is not None:
            lock.release_lock()
        return self.postprocess(self.response)

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.QUERY,
    )
    def query_qpu_config(self, host: str, server_port: int) -> QpuConfiguration:
        """Query the Qpu configuraiton of the target"""
        hostpath = f"{host}"
        if server_port:
            hostpath += f":{server_port}"
        response = requests.get(f"{hostpath}/qpu/config", timeout=10)
        qpu_config = QpuConfiguration()
        if response.ok:
            qpu_config.load_configuration(json.loads(response.text))
        return qpu_config



"""QPU Computation registry"""

import importlib
import os

from qstone.apps.PyMatching import PyMatching
from qstone.apps.QBC import QBC
from qstone.apps.RB import RB
from qstone.apps.VQE import VQE
from qstone.connectors import connector

# Mapping computation name to its class
_computation_registry = {
    "VQE": VQE,
    "RB": RB,
    "PyMatching": PyMatching,
    "QBC": QBC,
}

ENV_VARS = {
    "CONNECTIVITY_QPU_IP_ADDRESS": os.environ.get(
        "CONNECTIVITY_QPU_IP_ADDRESS", "127.0.0.1"
    ),
    "CONNECTIVITY_QPU_PORT": int(os.environ.get("CONNECTIVITY_QPU_PORT", "0")),
    "CONNECTIVITY_COMPILER_IP_ADDRESS": os.environ.get(
        "CONNECTIVITY_COMPILER_IP_ADDRESS", "127.0.0.1"
    ),
    "CONNECTIVITY_COMPILER_PORT": int(
        os.environ.get("CONNECTIVITY_COMPILER_PORT", "0")
    ),
    "CONNECTIVITY_MODE": connector.ConnectorType[
        os.environ.get("CONNECTIVITY_MODE", "NO_LINK")
    ],
    "QPU_MODE": os.environ.get("QPU_MODE", "RANDOM"),
    "CONNECTIVITY_TARGET": os.environ.get("CONNECTIVITY_TARGET", ""),
    "LOCKFILE": (
        os.environ.get("LOCK_FILE", "qstone.lock")
        if os.environ.get("SCHEDULING_MODE", "NONE") == "LOCK"
        else None
    ),
    "OUTPUT_PATH": os.environ.get("OUTPUT_PATH", ""),
    "JOB_ID": os.environ.get("JOB_ID", 0),
    "NUM_QUBITS": int(os.environ.get("NUM_QUBITS", "0")),
}


def get_computation_src(src: str):
    """Extracts the computation src either from the standard set or from the user"""
    if src in _computation_registry:
        computation_src = _computation_registry[src]  # type: ignore [abstract]
    else:
        try:
            src_module, src_class = src.rsplit(".", 1)
            computation_src = getattr(importlib.import_module(src_module), src_class)
        except NameError as exc:
            raise NameError(
                f"{src} app not found in standard list or as a folder"
            ) from exc
    return computation_src



"""Schema for the configuration file"""

FULL_SCHEMA = {
    "type": "object",
    "properties": {
        "environment": {
            "type": "object",
            "properties": {
                "project_name": {"type": "string"},
                "scheduling_mode": {"enum": ["LOCK", "SCHEDULER", "POLLING", "NONE"]},
                "lock_file": {"type": "string"},
                "job_count": {"type": "number"},
                "qpu": {
                    "type": "object",
                    "properties": {
                        "mode": {"enum": ["REAL", "EMULATED", "RANDOM"]},
                    },
                    "required": ["mode"],
                },
                "connectivity": {
                    "type": "object",
                    "properties": {
                        "mode": {"enum": ["NO_LINK", "HTTPS", "RIGETTI", "GRPC"]},
                        "ip_address": {"type": "string", "format": "hostname"},
                        "qpu": {
                            "type": "object",
                            "properties": {
                                "port": {"type": "number"},
                                "ip_address": {"type": "string", "format": "hostname"},
                            },
                        },
                        "compiler": {
                            "type": "object",
                            "properties": {
                                "port": {"type": "number"},
                                "ip_address": {"type": "string", "format": "hostname"},
                            },
                        },
                        "target": {"type": "string"},
                    },
                    "required": ["mode"],
                },
                "timeouts": {
                    "type": "object",
                    "properties": {
                        "http": {"type": "number"},
                        "lock": {"type": "number"},
                    },
                },
            },
            "required": ["scheduling_mode", "connectivity"],
            "jobs": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "qubit": {"type": "array", "minItems": 1, "maxItems": 2},
                        "shots": {"type": "array", "minItems": 1, "maxItems": 2},
                        "walltime": {"type": "number"},
                        "app_args": {"type": "object"},
                        "app_logging_level": {
                            "type": "integer",
                            "minimum": 0,
                            "maximum": 2,
                        },
                    },
                },
            },
            "users": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "user": {"type": "string"},
                        "job_count": {"type": "number"},
                        "computations": {"type": "object"},
                    },
                    "required": ["user", "computation"],
                },
            },
        },
    },
    "required": ["environment", "jobs", "users"],
}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qpu.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\tqpu.proto\x12\x16qstone.connectors.grpc"8\n\x07\x43ircuit\x12\x0f\n\x07\x63ircuit\x18\x01 \x01(\t\x12\x0e\n\x06pkt_id\x18\x02 \x01(\x05\x12\x0c\n\x04reps\x18\x03 \x01(\x05"3\n\x0f\x43ircuitResponse\x12\x0e\n\x06result\x18\x01 \x01(\t\x12\x10\n\x08\x63\x61pacity\x18\x02 \x01(\x05\x32\x66\n\x03QPU\x12_\n\x11RunQuantumCircuit\x12\x1f.qstone.connectors.grpc.Circuit\x1a\'.qstone.connectors.grpc.CircuitResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "qpu_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_CIRCUIT"]._serialized_start = 37
    _globals["_CIRCUIT"]._serialized_end = 93
    _globals["_CIRCUITRESPONSE"]._serialized_start = 95
    _globals["_CIRCUITRESPONSE"]._serialized_end = 146
    _globals["_QPU"]._serialized_start = 148
    _globals["_QPU"]._serialized_end = 250
# @@protoc_insertion_point(module_scope)



"""Wrapper for MPI calls"""

try:
    from qstone.multiprocessing.mpi import MPIHandler  # type: ignore [assignment]
except ModuleNotFoundError:
    from qstone.multiprocessing.nompi import MPIHandler  # type: ignore [assignment]



"""Quantum executor over a grpc channel"""

import json
import secrets

import grpc

import qstone.connectors.grpc.qpu_pb2 as pb2
import qstone.connectors.grpc.qpu_pb2_grpc as pb2_grpc
from qstone.connectors import connection
from qstone.utils.utils import ComputationStep, QpuConfiguration, trace


class GRPCConnecction(connection.Connection):
    """Connection running jobs over gRPC"""

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.PRE,
    )
    def preprocess(self, qasm_ptr: str) -> str:
        # Currently passthrough.
        with open(qasm_ptr, "r", encoding="utf-8") as fid:
            return fid.read()

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.POST,
    )
    def postprocess(self, message: str) -> str:
        # Currently passthrough.
        print(f"postprocess: {message}")
        return json.loads(message)

    # mypy: disable-error-code="attr-defined"
    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.RUN,
    )
    def run(
        self,
        qasm_ptr: str,
        reps: int,
        mode: str,
        qpu_host: str,
        qpu_port: int,
        compiler_host: str,
        compiler_port: int,
        target: str,
        lockfile: str,
    ) -> dict:
        compression = None  # grpc.Compression.None
        # instantiate a channel
        channel = grpc.insecure_channel(
            f"{qpu_host}:{qpu_port}", compression=compression
        )
        stub = pb2_grpc.QPUStub(channel)
        pkt_id = secrets.randbelow(2**31)
        circuit = self.preprocess(qasm_ptr)
        request = pb2.Circuit(circuit=circuit, pkt_id=pkt_id)  # type: ignore[attr-defined]
        m = stub.RunQuantumCircuit(request)
        return self.postprocess(m.result)

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.QUERY,
    )
    def query_qpu_config(self, qpu_host: str, qpu_port: int) -> QpuConfiguration:
        """Query the Qpu configuraiton of the target"""
        print(f"{qpu_host}:{qpu_port} pure printout - implemented!")
        return QpuConfiguration()



"""Profile utilities"""

import argparse
import logging
import os

import pandas as pd
import pandera.pandas as pa

from qstone.utils.utils import ComputationStep, load_json_profile, parse_json

PROFILER_SCHEMA = pa.DataFrameSchema(
    {
        "user": pa.Column(str),
        "prog_id": pa.Column(str),
        "job_id": pa.Column(str),
        "job_type": pa.Column(str),
        "job_step": pa.Column(
            str, checks=pa.Check.isin([e.name for e in ComputationStep])
        ),
        "start": pa.Column(int),
        "success": pa.Column(bool),
        "end": pa.Column(int),
    }
)


def _get_stats_from_dir(folder, schema):
    """
    Get the statistics from a folder applying the schema provided.
    """
    df = None
    for func_profile in os.listdir(folder):
        if func_profile.endswith(".json"):
            data = load_json_profile(os.path.join(folder, func_profile), schema)
            df = pd.concat([data, df], sort=False)
    logging.info("Folder: %s - found %d entries", folder, df.shape[0])
    return df


def _extrapolate(stats):
    """
    extrapolate provides an example of capabilities of Pandas.
    """
    # Adding an entry that defines the total duration
    stats["total"] = stats["end"] - stats["start"]
    # Aggregating micro-jobs with that belong to the same ID
    for pid in list(set(stats["prog_id"])):
        # Filtering per id
        mask = stats.prog_id == pid
        jobs = stats[stats.prog_id == pid]
        # We are aggregating all the steps associated with the same job
        for s in ["PRE", "RUN", "POST"]:
            stats.loc[mask, f"{s}_agg"] = jobs[jobs.job_step == s]["total"].sum()
    stats["count"] = len(stats[stats["success"]].groupby(["job_id", "user"]).groups)
    stats["connection_total"] = stats.query('job_type == "CONNECTION"')["total"].sum()
    return stats


def _store(stats, pickle):
    if os.path.exists(pickle):
        df = pd.concat([stats, pd.read_pickle(pickle)])
        df.to_pickle(pickle)
    else:
        stats.to_pickle(pickle)


NS_TO_MS = 1_000_000


def _print_stats(stats: pd.DataFrame):
    """
    Print general statistics
    """
    tot_classical = (stats["PRE_agg"].iloc[0] + stats["POST_agg"].iloc[0]) / NS_TO_MS
    tot_quantum = stats["RUN_agg"].iloc[0] / NS_TO_MS
    connection_total = stats["connection_total"].iloc[0] / NS_TO_MS
    tot_runs = stats["count"].iloc[0]
    print("########### Stats ######################")
    print(f"Total classical computation   [ms]:  {tot_classical:>12.2f}")
    print(f"Total quantum computation     [ms]:  {tot_quantum:>12.2f}")
    print(f"Average classical computation [ms]:  {tot_classical/tot_runs:>12.2f}")
    print(f"Average quantum computation   [ms]:  {tot_quantum/tot_runs:>12.2f}")
    print(f"Average connection time       [ms]:  {connection_total/tot_runs:>12.2f}")


def profile(
    config: str, folder: list[str], pickle: str
):  # pylint: disable=unused-argument
    """
    Profile the total execution across multiple users and store into
    a generalised pickled object.
    """
    # Get system configuration

    config_dict = parse_json(config)  # pylint: disable=unused-variable
    # Merging the results
    stats = pd.concat(
        [_get_stats_from_dir(f, PROFILER_SCHEMA) for f in folder], ignore_index=True
    )
    # Example of data extrapolation.
    _extrapolate(stats)
    # Store into an pickle file
    _store(stats, pickle)
    # Stats
    _print_stats(stats)


def main():
    """Main profile routine"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "cfg", type=str, help="Configuration file used to generate the load"
    )
    parser.add_argument("folder", type=str, help="Folder that contains the runs")
    parser.add_argument(
        "--pickle", type=str, help="Optional Pickle filepath to store pickled dataframe"
    )
    args = parser.parse_args()
    profile(args.config, args.folder, args.pickle)


if __name__ == "__main__":
    main()



"""QBC computations steps."""

import os

import numpy
from pandera import Check, Column, DataFrameSchema
from scipy.optimize import minimize

from qstone.apps.computation import Computation
from qstone.connectors import connector
from qstone.multiprocessing import MPIHandler
from qstone.utils.utils import ComputationStep, trace


@trace(
    computation_type="QBC",
    computation_step=ComputationStep.RUN,
    label="QASM_GENERATION",
)
def generate_vqc_qasm(pqc_number, num_qubits, datum, parameters):
    """
    Generates text for a qasm file of a Quantum Circuit (QC) consisting of encoding
    layers encoding a datum and Variational Quantum Circuit (VQC) with trainable
    parameters.
    The encoding layers are Rz and Rx layers at the beggining and end of the QC
    taking the datum as rotational angles.
    The VQC form can be chosen via pqc_number = {2, 5, 15}
    The pqc_number parameters references the QCs with the same number in
    figure 2 of Adv. Quantum Technol. 2019, 2, 1900070
    """

    qasm = 'OPENQASM 2.0;\ninclude "qelib1.inc";\nqreg q[{num_qubits}];\ncreg c[1];'
    for q in range(num_qubits):
        qasm += f"rx({datum[2*q]}) q[{q}]\n"
    for q in range(num_qubits):
        qasm += f"rz({datum[2*q+1]}) q[{q}]\n"

    if pqc_number == 2:
        for q in range(num_qubits):
            qasm += f"rx({parameters[q]}) q[{q}]\n"
        for q in range(num_qubits):
            qasm += f"rz({parameters[num_qubits+q]}) q[{q}]\n"
        for q in range(num_qubits - 1):
            qasm += f"CNOT q[{q}] q[{(q+1)}]\n"

    if pqc_number == 5:
        for q in range(num_qubits):
            qasm += f"rx({parameters[q]}) q[{q}]\n"
        for q in range(num_qubits):
            qasm += f"rz({parameters[num_qubits+q]}) q[{q}]\n"
        qasm += "gate crz(lambda) a,b\n{\nu1(lambda/2) b;\ncx a,b;\nu1(-lambda/2) b;\ncx a,b;\n}"
        i = 0
        for q in range(num_qubits):
            for qc in range(num_qubits):
                if q != qc:
                    qasm += f"crz({parameters[2*num_qubits+i]}) q[{q}] q[{(qc)}]\n"
                    i += 1
        for q in range(num_qubits):
            qasm += f"rx({parameters[(num_qubits+1)*num_qubits+q]}) q[{q}]\n"
        for q in range(num_qubits):
            qasm += f"rz({parameters[(num_qubits+2)*num_qubits+q]}) q[{q}]\n"

    if pqc_number == 15:
        for q in range(num_qubits):
            qasm += f"ry({parameters[q]}) q[{q}]\n"
        for q in range(num_qubits):
            qasm += f"CNOT q[{q}] q[{(num_qubits-1+q)%num_qubits}]\n"
        for q in range(num_qubits):
            qasm += f"ry({parameters[num_qubits+q]}) q[{q}]\n"
        for q in range(num_qubits):
            qasm += f"CNOT q[{(num_qubits-q)}] q[{(num_qubits+1-q)%num_qubits}]\n"

    # for q in range(num_qubits): qasm += f'rx({datum[2*q+1]}) q[{q}]\n'
    qasm += "measure q[0] -> c[0];\n"

    return qasm


@trace(
    computation_type="QBC",
    computation_step=ComputationStep.RUN,
    label="MPI_COMMUNICATION",
)
def mpi_communication(data, comm, bcast=True):
    """MPI communication wrapper"""
    return comm.communicate(data, bcast)


@trace(
    computation_type="QBC",
    computation_step=ComputationStep.RUN,
    label="LOSS_COMPUTATION",
)
def loss(
    parameters,
    pqc_number,
    num_qubits,
    shots,
    data,
    labels,
    idxs,
    comm,
    run_file,
    connection,
):
    """Loss function"""
    training_size = len(data)
    loss = 0

    mpi_communication(parameters, comm)

    for i in idxs:
        path = os.path.join(
            os.path.dirname(run_file), f"qbc_{os.environ['JOB_ID']}_{str(i)}.qasm"
        )

        datum = data[i]
        qasm = generate_vqc_qasm(pqc_number, num_qubits, datum, parameters)

        with open(path, "w", encoding="utf-8") as fid:
            fid.write(qasm)
        response = connection.run(qasm=path, reps=shots)
        if "counts" in response.keys():
            counts = response["counts"]
        else:
            print("ERROR: counts not found in qpu response.")
            quit()
            counts = response
        probs = {key: counts[key] / shots for key in counts.keys()}
        loss -= probs[str(labels[i])] / training_size
    print(f"partial loss for rank {comm.Get_rank()}: {loss}", flush=True)
    temp = mpi_communication(loss, comm, bcast=False)
    loss = temp
    print(f"total loss: {loss}", flush=True)

    return loss


class QBC(Computation):  # pylint:disable=invalid-name
    """
    QBC computation class.
    """

    COMPUTATION_NAME = "QBC"
    CFG_STRING = """
    {
      "cfg":
      {
        "num_required_qubits" : 4,
        "pqc_number" : 2,
        "training_size" : 20,
        "max_iters" : 10,
        "shots" : 64
      }
    }
    """
    SCHEMA = DataFrameSchema(
        {
            "pqc_number": Column(int, Check(lambda s: s >= 0)),
            "training_size": Column(int, Check(lambda s: s >= 0)),
            "max_iters": Column(int, Check(lambda s: s >= 0)),
            "shots": Column(int, Check(lambda s: s >= 0)),
        }
    )

    def __init__(self, cfg: dict):
        super().__init__(cfg)

        self.num_required_qubits: int
        self.repetitions: int
        self.training_size: int
        self.max_iters: int
        self.shots: int
        self.pqc_number: int

        self.qpu_cfg.num_required_qubits = self.num_required_qubits

        self.comm = MPIHandler()
        self.size = self.comm.Get_size()
        self.rank = self.comm.Get_rank()

    @trace(
        computation_type=COMPUTATION_NAME,
        computation_step=ComputationStep.PRE,
        label="DATA_GENERATION",
    )
    def pre(self, datapath: str):
        """Generates a random two-class dataset to train a Quantum Binary Classifier"""

        run_file = f"{datapath}/qbc_run_{os.environ['JOB_ID']}.npz"

        if self.rank == 0:
            data = numpy.pi * numpy.random.rand(
                self.training_size, 2 * self.num_required_qubits
            )
            labels = numpy.array(
                [0] * (self.training_size // 2 + self.training_size % 2)
                + [1] * (self.training_size // 2)
            )
            numpy.random.shuffle(labels)

            numpy.savez(run_file, data=data, labels=labels, allow_pickle=True)

        self.comm.Barrier()

    @trace(
        computation_type=COMPUTATION_NAME,
        computation_step=ComputationStep.RUN,
        label="MODEL_TRAINING",
    )
    def run(self, datapath: str, connection: connector.Connector):
        """Runs the VQC optimization"""

        run_file = f"{datapath}/qbc_run_{os.environ['JOB_ID']}.npz"

        model = numpy.load(run_file, allow_pickle=True)
        data = model["data"]
        labels = model["labels"]

        jobs_per_rank = self.training_size // self.size
        leftover = self.training_size % self.size
        if self.rank > self.size - leftover - 1:
            jobs_per_rank += 1
        jobsizes = self.comm.allgather(jobs_per_rank)  # type: ignore [attr-defined]
        starts = list(sum(jobsizes[:i]) for i in range(len(jobsizes)))
        idxs = numpy.arange(starts[self.rank], starts[self.rank] + jobsizes[self.rank])

        totqasms: list = []
        totresults: list = []
        totlosses: list = []
        if self.pqc_number in [2, 15]:
            totparameters = (
                2 * numpy.pi * numpy.random.rand(2 * self.num_required_qubits)
            )
        if self.pqc_number == 5:
            totparameters = (
                2
                * numpy.pi
                * numpy.random.rand(
                    pow(self.num_required_qubits, 2) + 3 * self.num_required_qubits
                )
            )

        mpi_communication(totparameters, self.comm)

        result = minimize(
            loss,
            totparameters,
            args=(
                self.pqc_number,
                self.num_required_qubits,
                self.shots,
                data,
                labels,
                idxs,
                self.comm,
                run_file,
                connection,
            ),
            method="COBYLA",
            options={"maxiter": self.max_iters},
        )
        print("x0", totparameters)
        for key in result.keys():
            print(key, result[key])
        store = {"data": data, "labels": labels, "result": result}
        if self.rank == 0:
            numpy.savez(run_file, **store, allow_pickle=True)

    @trace(computation_type=COMPUTATION_NAME, computation_step=ComputationStep.POST)
    def post(self, datapath: str):
        """Post process run result"""

        if self.rank == 0:
            run_file = f"{datapath}/qbc_run_{os.environ['JOB_ID']}.npz"
            print(f"Training results at {run_file}\n")

        self.comm.Barrier()  # type: ignore [attr-defined]



"""MPI wrapper module"""

from mpi4py import MPI  # pylint: disable=import-error


class MPIHandler:
    """MPIHandler class to wrap the MPI calls"""

    def __init__(self):
        self.comm = MPI.COMM_WORLD

    def communicate(self, data, bcast: bool):
        """communicate, transmits data in broadcast or singlecast"""
        if bcast:
            return self.comm.Bcast([data, MPI.DOUBLE], root=0)
        return self.comm.allreduce(data, op=MPI.SUM)

    def Get_size(self) -> int:  # pylint: disable=invalid-name
        """Redirects call to MPI Get_size"""
        return self.comm.Get_size()

    def Get_rank(self) -> int:  # pylint: disable=invalid-name
        """Redirects call to MPI Get_rank"""
        return self.comm.Get_rank()

    def Barrier(self):  # pylint: disable=invalid-name
        """Redirects call to MPI Barrier"""
        return self.comm.Barrier()

    def allgather(self, sendobj):  # pylint: disable=invalid-name
        """Redirects call to MPI allgather"""
        return self.comm.allgather(sendobj)



"""Quantum executor for Rigetti backend"""

import calendar
import time

import waiting
from pyquil import Program, get_qc

# pylint: disable=import-error,no-name-in-module
from qcs_sdk.compiler.quilc import (
    QuilcClient,
)
from qcs_sdk.qvm import QVMClient

# pylint: enable=import-error,no-name-in-module
from qstone.connectors import connection
from qstone.utils.utils import ComputationStep, trace


class RigettiConnection(connection.Connection):
    """Connection running jobs to Rigetti backend"""

    def __init__(self):
        """Creates the empty response"""
        self.response = None
        self.qc = None
        self.result = None
        self.mode = None
        self.origin = None

    def _get_qc(
        self,
        mode: str,
        qpu_host: str,
        qpu_port: int,
        compiler_host: str,
        compiler_port: int,
        target: str,
    ):
        self.mode = mode
        self.origin = qpu_host
        as_qvm = self.mode != "REAL"
        # compiler (non-standard port)
        # note: this a way to run a docker version of the compiler:
        # docker run --rm -it -p 5556:5556 rigetti/quilc -P -S -p 5556
        quilc_client = QuilcClient.new_rpcq(f"{compiler_host}:{compiler_port}")
        # qvm (non-standart port)
        # note, could run a docker version like this:
        # docker run --rm -it -p 5001:5001 rigetti/qvm -S  -p 5001
        qvm_client = QVMClient.new_http(f"{qpu_host}:{qpu_port}")
        return get_qc(
            target, as_qvm=as_qvm, quilc_client=quilc_client, qvm_client=qvm_client
        )

    def _run(self, program: Program):
        return self.qc.run(program)

    def _compile(self, circuit: Program, reps: int):
        return self.qc.compile(circuit.wrap_in_numshots_loop(reps))

    def _get_results(self, run):
        # "c" is the register used to store the classical value
        return run.data.result_data.to_register_map().get("c").to_ndarray()

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.PRE,
    )
    def preprocess(self, qasm_ptr: str) -> Program:
        """Preprocess the data. Transform QASM format into PyQuil"""
        with open(qasm_ptr, "r", encoding="utf-8") as fid:
            return self.qc.compiler.transpile_qasm_2(fid.read())

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.POST,
    )
    def postprocess(self, message: str) -> dict:
        """Postprocess the data"""
        # Data is returned in the format defined in assumptions.md.
        # Mapping is currently "complete" and in order.
        mapping = list(range(len(self.result[0])))
        timestamp = calendar.timegm(time.gmtime())
        origin = self.origin
        msg = {
            "mapping": mapping,
            "measurements": self.result,
            "mode": self.mode,
            "timestamp": timestamp,
            "origin": origin,
        }
        return msg

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.QUERY,
        label="_request_and_process",
    )
    def _request_and_process(
        self, qasm_ptr: str, reps: int, hostpath: str, lockfile: str
    ):
        circuit = self.preprocess(qasm_ptr)
        self.response = None
        success = False
        lock = connection.FileLock(lockfile)
        if lock.acquire_lock():
            run = self._run(self._compile(circuit, reps))
            self.result = self._get_results(run)
            success = True
        return success

    # mypy: disable-error-code="attr-defined"
    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.RUN,
    )
    def run(
        self,
        qasm_ptr: str,
        reps: int,
        mode: str,
        qpu_host: str,
        qpu_port: int,
        compiler_host: str,
        compiler_port: int,
        target: str,
        lockfile: str,
    ) -> dict:
        """Run the connection to the server"""
        self.qc = self._get_qc(
            mode, qpu_host, qpu_port, compiler_host, compiler_port, target
        )
        try:
            waiting.wait(
                lambda: self._request_and_process(qasm_ptr, reps, qpu_host, lockfile),
                timeout_seconds=20,
            )
        except waiting.TimeoutExpired:
            pass
        return self.postprocess(self.response)



"""Connection abstraction for nodes."""

import os
from abc import ABC, abstractmethod
from pathlib import Path


class FileLock:
    """Implementation of the file locking system."""

    def __init__(self, lockfile: str):
        """Lock creation"""
        self._lockfile = lockfile

    def acquire_lock(self) -> bool:
        """Tries to acquire the lock. Atomic way"""
        if self._lockfile is None:
            locked = True
        else:
            try:

                # Atomic operation
                Path(self._lockfile).touch(exist_ok=False)
                locked = True
            except FileExistsError:
                locked = False
        return locked

    def release_lock(self):
        """Releases the lock."""
        if self._lockfile:
            os.remove(self._lockfile)


class Connection(ABC):
    """Abstract class to represent connection between nodes"""

    @abstractmethod
    def preprocess(self, qasm_ptr: str) -> str:
        """Preprocess the data."""

    @abstractmethod
    def postprocess(self, message: str) -> dict:
        """Postprocess the data"""

    @abstractmethod
    def run(
        self,
        qasm_ptr: str,
        reps: int,
        mode: str,
        qpu_host: str,
        qpu_port: int,
        compiler_host: str,
        compiler_port: int,
        target: str,
        lockfile: str,
    ) -> dict:
        """Run the connection to the server"""



"""General utilities. Used across the jobs"""

import json
import os
import random
import re
import time
from enum import Enum
from functools import wraps
from typing import Callable, Dict, Optional

import jsonschema
import pandas as pd
import pandera.pandas as pa

from .config_schema import FULL_SCHEMA


class JobReturnCode(Enum):
    """Type of exit code of job run"""

    JOB_COMPLETED = 0
    INSUFFICIENT_QPU_RESOURCES = 1
    PRE_STEP_INCOMPLETE = 2
    RUN_STEP_INCOMPLETE = 3
    POST_STEP_INCOMPLETE = 4


class ComputationStep(Enum):
    """Step of computation for profiling"""

    PRE = "PRE"
    RUN = "RUN"
    POST = "POST"
    QUERY = "QUERY"


CFG_ENVIRONMENT_VARIABLES = {
    "project_name",
    "connector",
    "qpu_ip_address",
    "qpu_port",
    "qpu_management",
    "lock_file",
    "timeouts.http",
    "timeouts.lock",
}


def validate_computation_weights(config: Dict):
    users = config["users"]  #
    for user in users:
        weights_sum = sum(user["computations"].values())
        if abs(weights_sum - 1.0) > 1e-6:  # Using epsilon for float comparison
            raise ValueError(
                f"Sum of computation weights for user {user['user']} is {weights_sum}, not 1.0"
            )


def parse_json(config: str) -> Dict:
    """
    Parses the JSON file, validates it against the schema and returns a dictionary
    representation
    """
    with open(config, "r", encoding="UTF-8") as f:
        config_dict = json.loads(f.read())
    jsonschema.validate(config_dict, FULL_SCHEMA)
    validate_computation_weights(config_dict)
    return config_dict


class QpuConfiguration:
    """Defines the configuration of a QPU (Quantum Processing Unit)"""

    def __init__(self) -> None:
        """Set Qpu Configuration defaults"""
        self.qpu_ip_address = "0"
        self.qpu_port = "0"

    def load_configuration(self, config: dict) -> None:
        """Loads QPU configuration data"""
        self.qpu_ip_address = config["connectivity"]["qpu"]["ip_address"]
        self.qpu_port = config["connectivity"]["qpu"]["port"]

    def write_configuration(self, output_path: str) -> None:
        """Writes QPU configuration as json

        Args:
            output_path: File path to write configuration to
        """
        config = {"QPU_ADDRESS": self.qpu_ip_address, "QPU_PORT": self.qpu_port}
        with open(output_path, "w", encoding="utf-8") as file:
            json.dump(config, file)


def qasm_circuit_random_sample(qasm: str, repetitions: int) -> Dict:
    """Mocks simulation of qasm circuit by giving random readouts for classical registers

    Args:
        qasm: string representation of qasm circuit
        repetitions: number of readouts to simulate
    Returns frequency of each classical bit string sampled
    """
    # Extract number classical registers
    creg_defs = re.findall(r"creg [a-zA-Z]\w*\[\d+\]", qasm)
    num_cregs = 0
    for c in creg_defs:
        num_cregs += int(re.findall(r"\d+", c)[0])

    # Extract the mapping between quantum and classical registers
    mapping = []
    qreg_meas = re.findall(r"[a-zA-Z]\w*\[\d+\] -> ", qasm)
    for qreg in qreg_meas:
        trimmed = qreg[len("q[")]
        mapping.append(int(re.findall(r"\d+", trimmed)[0]))

    # Generate a random readout per shot
    measurements = []
    counts: dict = {}
    for _ in range(repetitions):
        meas = list(list(random.randint(0, 1) for _ in range(num_cregs)))
        key = "".join(str(bit) for bit in meas)
        measurements.append(meas)
        if key not in counts.keys():
            counts[key] = 1
        else:
            counts[key] += 1

    return {
        "mapping": mapping,
        "measurements": measurements,
        "counts": counts,
        "mode": "random source",
    }


def _get_job_id():
    """Returns the job id from the tool"""
    return os.environ["JOB_ID"]


def _get_content(
    times: tuple[int, int],
    computation_type: str,
    computation_step: ComputationStep,
    label: Optional[str],
    success: bool,
):
    """Returns the dictionary that represents the time trace datapoint"""
    content = {}
    content["user"] = os.environ["QS_USER"]
    content["prog_id"] = os.environ["PROG_ID"]
    content["job_id"] = _get_job_id()
    content["job_type"] = computation_type
    content["job_step"] = computation_step.value
    content["label"] = label  # type: ignore[assignment]
    content["start"] = times[0]  # type: ignore[assignment]
    content["end"] = times[1]  # type: ignore[assignment]
    content["success"] = success  # type: ignore[assignment]
    return content


def _write_trace(
    profile_path: str,
    times: tuple[int, int],
    computation_type: str,
    computation_step: ComputationStep,
    label: str,
    success: bool,
):
    """Handler for trace writing"""
    trace_content = _get_content(
        times, computation_type, computation_step, label, success
    )
    with open(profile_path, "w", encoding="utf-8") as fid:
        json.dump(trace_content, fid, ensure_ascii=False, indent=4)


def trace(
    computation_type: str,
    computation_step: ComputationStep,
    label: Optional[str] = None,
    logging_level: Optional[int] = 2,
):
    """General tracing of the function. Wrapper"""

    def wrapper(func: Callable):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            logging_level_met = logging_level >= int(
                os.environ.get("APP_LOGGING_LEVEL", "0")
            )
            start = time.perf_counter_ns()
            profile_name = "_".join(
                filter(
                    None,
                    (
                        "job",
                        _get_job_id(),
                        computation_step.value,
                        computation_type,
                        label,
                        str(start),
                    ),
                )
            )
            profile_path = os.path.join(
                os.environ["PROFILE_PATH"], f"{profile_name}.json"
            )
            try:
                result = func(*args, **kwargs)
                success = True
            except Exception as e:  # pylint: disable=broad-except
                result = None
                success = False
                raise e
            finally:
                end = time.perf_counter_ns()
                if logging_level_met:
                    _write_trace(
                        profile_path,
                        (start, end),
                        computation_type,
                        computation_step,
                        label,
                        success,
                    )
            return result

        return wrapper_func

    return wrapper


def load_json_profile(trace_info: str, schema: pa.DataFrameSchema) -> pd.DataFrame:
    """Loads function json profile and checks it against the schema

    Args:
        trace_info: File location of function traced information
        schema: Validator schema

    Returns pandas dataframe containing profile information
    """
    with open(trace_info, "r", encoding="utf-8") as f:
        df = pd.json_normalize(json.load(f))
    schema.validate(df)
    return df



"""Runner for no-link connector"""

from qstone.connectors import connection
from qstone.utils.utils import (
    ComputationStep,
    qasm_circuit_random_sample,
    trace,
)


class NoLinkConnection(connection.Connection):
    """No link connection running jobs without a server"""

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.PRE,
    )
    def preprocess(self, qasm_ptr: str) -> str:
        """Read QASM circuit files"""
        with open(qasm_ptr, "r", encoding="utf-8") as fid:
            qasm = fid.read()
        return qasm

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.PRE,
    )
    def postprocess(self, message: str) -> dict:
        # Currently passthrough.
        # generate one random entry with correct number of shots
        # REVISIT(mghibaudi, "use self.num_qubits to generate the values")
        return {"00": 0, "01": 0, "10": 0, "11": 0}

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.POST,
        label="get_outcomes",
    )
    def _get_outcomes(self, qasm_circuit: str, reps: int) -> dict:
        return qasm_circuit_random_sample(qasm_circuit, reps)

    @trace(
        computation_type="CONNECTION",
        computation_step=ComputationStep.POST,
    )
    def run(
        self,
        qasm_ptr: str,
        reps: int,
        mode: str,
        qpu_host: str,
        qpu_port: int,
        compiler_host: str,
        compiler_port: int,
        target: str,
        lockfile: str,
    ) -> dict:
        """Local simulated run of circuit"""
        qasm_circuit = self.preprocess(qasm_ptr)
        outcomes = qasm_circuit_random_sample(qasm_circuit, reps)
        return outcomes
